package edu.njit.decaf3;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.List;

import edu.njit.decaf3.helper.Control;
import edu.njit.decaf3.helper.JSON;
import javolution.util.FastList;
import javolution.util.FastTable;

import edu.njit.decaf3.structure.Simulation;
import org.jblas.DoubleMatrix;
import org.jblas.Solve;

/**
 * DECAF - Dependability Evaluator of Cascading Failures Revision 3
 *
 * @author Sashank Tadepalli, Mihir Sanghavi
 */
public class DECAF {
	// Required class fields
	private Simulation sim;
	public double[][] qMatrix;
	public int length = 1;
	public int upLength = 1;
	private int envLength;
	private int nodeLength;
	private FastList<Integer> upStates;

	// Approximation parameters
	public int heightThreshold = 0;
	public double rateThreshold = 0;
	public int treeThreshold = 0;
	public boolean stateThreshold = false;

	// Data caches
	private List<int[]>[] binaryEnumCache;
	private int rootIndex = -1;

	// Statistics
	private long t;
	public int numberOfTrees;
	public int numberOfTransitions;
	public double nonFailureTime;
	public double failureTransitionTime;
	public double MTTFSolveTime;
	public double SSUSolveTime;

	/**
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	public DECAF() throws Exception {
		this(new File("data/simulation.model.json"));
	}

	/**
	 * @param inputFileLocation
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	public DECAF(String inputFileLocation) throws Exception {
		this(new File(inputFileLocation));
	}

	/**
	 * @param inputFile
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	public DECAF(File inputFile) throws Exception {
		// Assert inputFile exists
		if (!inputFile.exists()) {
			throw new FileNotFoundException(inputFile.getPath() + " cannot be found in " +
					new File(".").getCanonicalPath());
		}

		t = System.nanoTime();
		// Load Json Data
		sim = JSON.parse(inputFile);
		init();
		nonFailureTime += (System.nanoTime() - t) / 1e9D;
	}

	/**
	 * @param s
	 */
	public DECAF(Simulation s) {
		sim = s;
		init();
	}

	private void init(){
		initLength();
		initUpStates();
		initCascading();
		initBEC();
	}

	/**
	 *
	 */
	public void run() {
		t = System.nanoTime();

		// Initialize QMatrix
		qMatrix = new double[length][length];

		nonFailureTime += (System.nanoTime() - t) / 1e9D;

		// Fill QMatrix
		t = System.nanoTime();
		fillEnvChangeTransitions();
		nonFailureTime += (System.nanoTime() - t) / 1e9D;

		t = System.nanoTime();
		fillRepairTransitions();
		nonFailureTime += (System.nanoTime() - t) / 1e9D;

		t = System.nanoTime();
		fillFailureTransitions();
		failureTransitionTime += (System.nanoTime() - t) / 1e9D;
	}

	private void initLength() {
		if (sim == null)
			Control.end("Simulation is not initialized, please refer to manual for usage.");

		nodeLength = sim.components.length;
		envLength = sim.environments.length;
		length = envLength;
		for (int i = 0; i < nodeLength; i++)
			length *= sim.components[i].redundancy + 1;
	}

	private void initUpStates() {
		if (sim == null)
			Control.end("Simulation is not initialized, please refer to manual for usage.");

		upStates = new FastList<>();

		// Calculate up states (if not up, ignore from P_R matrix)
		here:
		for (int i = 0; i < length; i++) {
			int[] tempState = intToState(i);
			for (int k = 0; k < nodeLength; k++) {
				if (sim.components[k].redundancy - tempState[k] < sim.components[k].required) {
					continue here;
				}
			}
			upStates.add(i);
		}
		upLength = upStates.size();
	}

	private void initCascading() {
		if (sim == null || nodeLength == 0)
			Control.end("Simulation is not initialized, please refer to manual for usage.");

		for (int i = 0; i < nodeLength; i++)
			for (int j = 0; j < nodeLength; j++)
				if (sim.cascading[i][j] != 0) sim.components[i].cascading.add(j);

	}

	private void initBEC(){
		binaryEnumCache = new FastTable[nodeLength];
		for (int i = 0; i < nodeLength; i++) {
			if (sim.components[i].cascading.isEmpty()) continue;
			binaryEnumCache[i] = powerSet(sim.components[i].cascading);
		}
	}

	/**
	 *
	 */
	private void fillEnvChangeTransitions() {
		// Iterate |qMatrix| skipping over |environments|
		for (int i = 0; i < length; i += envLength) {

			// Overlay the environment change matrix
			for (int j = 0; j < envLength; j++) {
				for (int k = 0; k < envLength; k++) {
					qMatrix[i + j][i + k] = sim.environments[j][k];
				}
			}
		}
	}

	/**
	 *
	 */
	private void fillRepairTransitions() {
		// Iterate |qMatrix|
		for (int i = 0; i < length; i++) {

			// Increment each node and calculate the resulting to-state
			for (int j = 0; j < nodeLength; j++) {
				int[] currState = intToState(i);
				currState[j] += 1;
				int toIndex = stateToInt(currState);

				if (toIndex < 0) continue;

				int sum = 0;
				for (int k = 0; k < currState.length - 1; k++) {
					sum += currState[k];
				}

				// Populate respective qMatrix cell
				// with (number of nodes in from-state) * (repair-rate of node)
				// / (sum of nodes in from-state)
				qMatrix[toIndex][i] = currState[j] * sim.components[j].repair[currState[nodeLength]] / sum;
			}
		}
	}

	/**
	 *
	 */
	private void fillFailureTransitions() {
		for (int i = 0; i < nodeLength; i++) {
			rootIndex = i;

			int[] initialTransition = intToState(0);
			initialTransition[i] = 1;

			double[][] bfhMap = buildBFH(null);
			int[] bfhCounter = new int[sim.components.length];
			bfhMap[i][0] = 1;
			bfhMap[i][1] = 1;
			bfhCounter[i]++;

			growSubTree(new int[]{i}, initialTransition, bfhMap, bfhCounter, 1F, 0);
		}
	}

	/**
	 * @param currentLevel
	 * @param transition
	 * @param subTreeRate
	 * @param bfhMap
	 */
	private void growSubTree(int[] currentLevel, int[] transition, double[][] bfhMap, int[] bfhCounter,
	                         double subTreeRate, int heightCounter) {
		int[] gammaPermutations = new int[0];
		int[] terminalTypes = new int[0];
		boolean first = true;

		for (int t = 0; t < currentLevel.length; t++) {
			if (currentLevel[t] >= 0 && binaryEnumCache[currentLevel[t]] != null) {
				terminalTypes = APACHE_ARRAYUTILS_addAll(terminalTypes, new int[]{currentLevel[t]});
				gammaPermutations = APACHE_ARRAYUTILS_addAll(gammaPermutations,
						new int[]{binaryEnumCache[currentLevel[t]].size()});
			}
		}
		List<int[]> productSet = new FastTable<>();
		cartesianProduct(gammaPermutations, 0, new int[0], productSet);
		SKIPLEVEL:
		for (int[] oneNextLevelPossibility : productSet) {
			int[] ftCopy = transition.clone();
			double[][] bfhCopy = buildBFH(bfhMap);
			int[] bfhCounterCopy = bfhCounter.clone();
			double subTreeRateCopy = subTreeRate;
			int[] newLevel = new int[0];

			for (int b = 0; b < oneNextLevelPossibility.length; b++) {
				int parentType = terminalTypes[b];
				int binEnumId = oneNextLevelPossibility[b];
				int[] block = binaryEnumCache[parentType].get(binEnumId);
				newLevel = APACHE_ARRAYUTILS_addAll(newLevel, block);
				for (int childType : block) {
					if (childType >= 0) {
						if (ftCopy[childType] >= sim.components[childType].redundancy) continue SKIPLEVEL;

						bfhCopy[childType][bfhCounterCopy[childType] + 1] = bfhCopy[childType][bfhCounterCopy[childType]];
						bfhCounterCopy[childType]++;
						ftCopy[childType]++;
						subTreeRateCopy *= sim.cascading[parentType][childType];
					} else {
						// Not failed
						childType = Math.abs(childType) - 1;
						bfhCopy[childType][bfhCounterCopy[childType]] *= 1 - sim.cascading[parentType][childType];
					}
				}
			}

			if (treeThreshold > 0) {
				int sum = 0;
				for (int i = 0; i < transition.length; i++) {
					sum += transition[i];
					if (sum > treeThreshold) continue SKIPLEVEL;
				}
			}

			if (stateThreshold) {
				for(int i = 0; i < nodeLength; i++)
					if(transition[i] > sim.components[i].redundancy - sim.components[i].required)
						continue SKIPLEVEL;
			}

			if (heightThreshold > 0 && heightCounter > heightThreshold) continue SKIPLEVEL;

			if (subTreeRate < rateThreshold) continue SKIPLEVEL;

			if (first) {
				processRates(transition, bfhCopy, bfhCounterCopy, subTreeRate);
				first = false;
			} else {
				growSubTree(newLevel, ftCopy, bfhCopy, bfhCounterCopy, subTreeRateCopy, heightCounter + 1);
			}
		}
	}

	/**
	 * @param transition
	 * @param bfHist
	 * @param bfhCounter
	 * @param subTreeRate
	 */
	private void processRates(int[] transition, double[][] bfHist, int[] bfhCounter, double subTreeRate) {
		numberOfTrees++;
		here:
		for (int i = 0; i < length; i += envLength) {
			int[] toState = new int[nodeLength + 1];
			int[] fromState = intToState(i);
			for (int k = 0; k < nodeLength; k++) {
				toState[k] = fromState[k] + transition[k];
				if (toState[k] > sim.components[k].redundancy) {
					continue here;
				}
			}
			int toIndex = stateToInt(toState);

			double rootRate = (sim.components[rootIndex].redundancy - fromState[rootIndex]);
			double complementRate = 1D;

			for (int k = 0; k < nodeLength; k++) {
				int compsAvailable = sim.components[k].redundancy - fromState[k];
				if (compsAvailable <= 0) {
					continue;
				}
				complementRate *= bfHist[k][Math.min(compsAvailable - 1, bfhCounter[k])];
			}
			for (int j = 0; j < envLength; j++, toIndex++) {
				numberOfTransitions++;
				qMatrix[i + j][toIndex] += rootRate * subTreeRate
						* complementRate * sim.components[rootIndex].failure[j];
			}
		}
	}

	/**
	 * Index that state can be found in QMatrix, -1 otherwise
	 *
	 * @param currState
	 * @return
	 */
	private int stateToInt(int[] currState) {
		int result = 0;
		int temp = 1;
		for (int i = nodeLength - 1; i >= 0; i--) {
			if (currState[i] < 0 || currState[i] > sim.components[i].redundancy) {
				return -1;
			}
			result += currState[i] * temp;
			temp *= sim.components[i].redundancy + 1;
		}
		result = result * envLength + currState[nodeLength];
		return result;
	}

	/**
	 * [ A, B, e ]
	 *
	 * @param i
	 * @return
	 */
	public int[] intToState(int i) {
		if (i > length || i < 0) return null;

		int[] result = new int[nodeLength + 1];
		result[result.length - 1] = i % envLength;
		i /= envLength;
		for (int j = nodeLength - 1; j >= 0; j--) {
			result[j] = i % (sim.components[j].redundancy + 1);
			i = (i - result[j]) / (sim.components[j].redundancy + 1);
		}
		return result;
	}

	/*
	 * private String stateToString(int[] state) { String result = ""; for (int i = 0; i < nodeLength; i++) {
	 * result += ", " + state[i]; } return "(" + result.substring(2) + ", " + state[nodeLength] + ")"; }
	 */

	private static List<int[]> powerSet(List<Integer> componentList) {
		FastTable<Integer> members = new FastTable<Integer>(componentList);
		if (members.isEmpty()) {
			return new FastTable<int[]>();
		}

		int permutations = (int) Math.pow(2, members.size());
		FastTable<int[]> binaryEnum = new FastTable<int[]>(permutations);
		for (int p = 0; p < permutations; p++) {
			StringBuffer binary = new StringBuffer(String.format("%" + members.size() + "s", Integer.toBinaryString(p))
					.replace(' ', '0'));
			int[] block = new int[binary.length()];
			for (int b = 0; b < binary.length(); b++) {
				block[b] = members.get(b) - (binary.charAt(b) == '0' ? members.get(b) * 2 + 1 : 0);
			}
			binaryEnum.add(p, block);
		}
		return binaryEnum;
	}

	/**
	 * @param copy
	 * @return
	 */
	private double[][] buildBFH(double[][] copy) {
		double[][] result = new double[sim.components.length][];
		for (int i = 0; i < nodeLength; i++) {
			if (copy == null) {
				result[i] = new double[sim.components[i].redundancy + 1];
				result[i][0] = 1;
			} else {
				result[i] = copy[i].clone();
			}
		}
		return result;
	}

	/**
	 * @param limits
	 * @param index
	 * @param current
	 * @param list
	 */
	private void cartesianProduct(int[] limits, int index, int[] current, List<int[]> list) {
		if (current.length == limits.length) {
			list.add(current);
		}
		if (index >= limits.length) {
			return;
		}
		for (int i = 0; i < limits[index]; i++) {
			cartesianProduct(limits, index + 1, APACHE_ARRAYUTILS_addAll(current.clone(), new int[]{i}), list);
		}
	}

	/**
	 * Taken from Apache Commons Library - ArrayUtils.addAll
	 *
	 * @param array1
	 * @param array2
	 * @return
	 */
	public static int[] APACHE_ARRAYUTILS_addAll(int[] array1, int[] array2) {
		if (array1 == null) return APACHE_ARRAYUTILS_clone(array2);
		else if (array2 == null) return APACHE_ARRAYUTILS_clone(array1);
		int[] joinedArray = new int[array1.length + array2.length];
		System.arraycopy(array1, 0, joinedArray, 0, array1.length);
		System.arraycopy(array2, 0, joinedArray, array1.length, array2.length);
		return joinedArray;
	}

	/**
	 * Taken from Apache Commons Library - ArrayUtils.clone
	 *
	 * @param array
	 * @return
	 */
	public static int[] APACHE_ARRAYUTILS_clone(int[] array) {
		if (array == null) return null;
		return array.clone();
	}

	/**
	 *
	 */
	public double getMTTF() {
		long t = System.nanoTime();
		prepareQMatrix();
		int subMatrixLength = upStates.size();
		double[] hVector = new double[subMatrixLength];
		double[][] pMatrix = new double[subMatrixLength][subMatrixLength];
		for (int i = 0; i < subMatrixLength; i++) {
			hVector[i] = -qMatrix[upStates.get(i)][upStates.get(i)];
			for (int j = 0; j < subMatrixLength; j++) {
				pMatrix[i][j] = qMatrix[upStates.get(i)][upStates.get(j)] / -hVector[i];
			}
		}
		double result = 0.0;
		DoubleMatrix temp = Solve.pinv(new DoubleMatrix(pMatrix));
		for (int i = 0; i < subMatrixLength; i++) {
			result += hVector[i] * temp.get(0, i);
		}
		MTTFSolveTime += (System.nanoTime() - t) / 1e9D;
		return result;
	}

	public double getSSU() {
		long t = System.nanoTime();

		double[][] qPrimeMatrix = qMatrix.clone();
		double result = 1.0;
		for (int i = 0; i < length; i++) {
			qPrimeMatrix[i][0] = 1;
		}
		DoubleMatrix temp = Solve.pinv(new DoubleMatrix(qPrimeMatrix));
		for (Integer k : upStates) {
			result -= temp.get(0, k);
		}
		SSUSolveTime += (System.nanoTime() - t) / 1e9D;
		return result;
	}

	private void prepareQMatrix() {
		for (int i = 0; i < qMatrix.length; i++) {
			qMatrix[i][i] = 0;
			double sum = 0;
			for (int j = 0; j < qMatrix[i].length; j++) {
				sum += qMatrix[i][j];
			}
			qMatrix[i][i] = -sum;
		}
	}
}